#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import calendar
import enum
import math
import time
from datetime import datetime, timedelta, timezone

from construct import (
    Adapter,
    BitsInteger,
    BytesInteger,
    Container,
    Flag,
    Float32l,
    Int8ul,
    Int16sl,
    Int16ul,
    Int24ul,
    Padding,
    StopIf,
    Struct,
    this,
)

from bluetooth_mesh.messages.util import (
    EmbeddedBitStruct,
    EnumAdapter,
    EnumSwitch as Switch,
    NamedSelect,
    Opcode,
    SwitchStruct,
)

MS_IN_UNCERTAINTY_STEP = 10
UNCERTAINTY_MS = 10
CURRENT_TAI_UTC_DELTA = 37
TAI_UTC_DELTA_ZERO = 0xFF
TIME_ZONE_OFFSET_ZERO = 0x40
MESH_UNIX_EPOCH_DIFF = calendar.timegm(
    time.strptime("2000-01-01T00:00:00", "%Y-%m-%dT%H:%M:%S")
) - calendar.timegm(time.gmtime(0))
SECONDS_IN_15_MINUTES = 15 * 60


# fmt: off


class TimeRole(enum.IntEnum):
    NONE = 0
    TIME_AUTHORITY = 1
    TIME_RELAY = 2
    TIME_CLIENT = 3


def TAI_UTC_DeltaPaddedField(tai_utc_delta_name: str):
    return EmbeddedBitStruct("_",
                             Padding(1),
                             tai_utc_delta_name / BitsInteger(15),
                             reversed=True
                             )


def mesh_time_zone_offset_to_timedelta(time_zone_offset: int) -> timedelta:
    return timedelta(seconds=(time_zone_offset - TIME_ZONE_OFFSET_ZERO) * SECONDS_IN_15_MINUTES)


def timedelta_to_mesh_time_zone_offset(time_zone: timedelta) -> int:
    if time_zone is None:
        return TIME_ZONE_OFFSET_ZERO

    assert (int(time_zone.total_seconds() % SECONDS_IN_15_MINUTES) == 0)
    return int((time_zone.total_seconds() // SECONDS_IN_15_MINUTES) + TIME_ZONE_OFFSET_ZERO)


def mesh_tai_utc_delta_to_timedelta(tai_utc_delta: int) -> timedelta:
    return timedelta(seconds=tai_utc_delta - TAI_UTC_DELTA_ZERO)


def timedelta_to_mesh_tai_utc_delta(time_zone: timedelta) -> int:
    assert (time_zone.total_seconds().is_integer())
    return int(time_zone.total_seconds() + TAI_UTC_DELTA_ZERO)


def subsecond_to_seconds(subsecond: int) -> float:
    return subsecond / 256


def seconds_to_subsecond(seconds: float) -> int:
    return round((seconds - int(seconds)) * 256)

TimeMinimal = Struct(
    "tai_seconds" / BytesInteger(5, swapped=True),
)

TimeOptional = Struct(
    "tai_seconds" / BytesInteger(5, swapped=True),
    StopIf(this.tai_seconds == 0),
    "subsecond" / Int8ul,
    "uncertainty" / Int8ul,
    *EmbeddedBitStruct("_",
                       "tai_utc_delta" / BitsInteger(15),
                       "time_authority" / Flag,
                       reversed=True
                       ),
    "time_zone_offset" / Int8ul,
)

Time = NamedSelect(
    optional=TimeOptional,
    minimal=TimeMinimal,
)


class TimeZoneOffsetAdapter(Adapter):
    """
    Time Zone Offset is described in minutes, mesh format in 15-minute increments.
    """

    def _decode(self, obj, context, path):
        return (obj - 0x40) * 15

    def _encode(self, obj, context, path):
        assert ((obj % 15) == 0)
        return 0x40 + (obj // 15)


class TAIUTCDeltaAdapter(Adapter):
    """
    TAI-UTC Delta is described in seconds encoded in signed integer, mesh format in signed integer with different offset
    """

    def _decode(self, obj, context, path):
        return obj - 0xFF

    def _encode(self, obj, context, path):
        assert (obj > -255)
        return obj + 0xFF


class UncertaintyAdapter(Adapter):
    """
    TAI-UTC Delta is described in seconds encoded in float, mesh format is in centiseconds
    """

    def _decode(self, obj, context, path):
        return round(obj / 100, 2)

    def _encode(self, obj, context, path):
        assert (obj < 2.6)
        return math.floor(obj * 100)


DateTime = Struct(
    "year" / Int16ul,
    "month" / Int8ul,
    "day" / Int8ul,
    "hour" / Int8ul,
    "minute" / Int8ul,
    "second" / Int8ul,
    "microsecond" / Int24ul,
    "time_zone_offset" / TimeZoneOffsetAdapter(Int16sl),
)


class TimeAdapter(Adapter):
    _subcon = Struct(
        "date" / DateTime,
        "tai_utc_delta" / TAIUTCDeltaAdapter(Int16sl),
        "time_authority" / Flag,
        "uncertainty" / UncertaintyAdapter(Float32l),
    )

    def _decode(self, obj, context, path):
        if obj["tai_seconds"] == 0:
            return Container(
                date=None,
                tai_utc_delta=None,
                time_authority=None,
                uncertainty=None
            )
        time_zone = mesh_time_zone_offset_to_timedelta(obj["time_zone_offset"])
        full_recv_time = obj["tai_seconds"] + subsecond_to_seconds(obj["subsecond"]) + MESH_UNIX_EPOCH_DIFF + int(
            time_zone.total_seconds())
        recv_date = datetime.fromtimestamp(full_recv_time, timezone(time_zone))

        return Container(
            date=recv_date,
            tai_utc_delta=mesh_tai_utc_delta_to_timedelta(obj["tai_utc_delta"]),
            time_authority=bool(obj["time_authority"]),
            uncertainty=timedelta(milliseconds=(obj["uncertainty"] * 10))
        )

    def _encode(self, obj, context, path):
        passed_time: datetime = obj["date"]

        if isinstance(passed_time, dict):  # capnproto message
            time_zone = timedelta(minutes=passed_time["time_zone_offset"])
            passed_time = datetime(
                year=passed_time["year"],
                month=passed_time["month"],
                day=passed_time["day"],
                hour=passed_time["hour"],
                minute=passed_time["minute"],
                second=passed_time["second"],
                microsecond=passed_time["microsecond"],
                tzinfo=timezone(time_zone)
            )

        if isinstance(obj["uncertainty"], float):
            obj["uncertainty"] = timedelta(seconds=obj["uncertainty"])

        if isinstance(obj["tai_utc_delta"], int):
            obj["tai_utc_delta"] = timedelta(seconds=obj["tai_utc_delta"])

        total_time = passed_time.timestamp() - MESH_UNIX_EPOCH_DIFF - passed_time.utcoffset().total_seconds()

        return Container(
            tai_seconds=int(total_time),
            subsecond=seconds_to_subsecond(total_time),
            uncertainty=int((obj["uncertainty"].total_seconds() * 100)),
            tai_utc_delta=timedelta_to_mesh_tai_utc_delta(obj["tai_utc_delta"]),
            time_authority=bool(obj["time_authority"]),
            time_zone_offset=timedelta_to_mesh_time_zone_offset(passed_time.utcoffset())
        )


TimeRoleMsg = Struct(
    "time_role" / EnumAdapter(Int8ul, TimeRole)
)

TimeGet = Struct()

TimeSet = TimeAdapter(Time)

TimeStatus = TimeAdapter(Time)

TimeZoneGet = Struct()

TimeZoneSet = Struct(
    "time_zone_offset_new" / Int8ul,
    "tai_of_zone_change" / BytesInteger(5, swapped=True),
)

TimeZoneStatus = Struct(
    "time_zone_offset_current" / Int8ul,
    "time_zone_offset_new" / Int8ul,
    "tai_of_zone_change" / BytesInteger(5, swapped=True),
)

TAIUTCDeltaGet = Struct()

TAIUTCDeltaSet = Struct(
    *TAI_UTC_DeltaPaddedField("tai_utc_delta_new"),
    "tai_of_delta_change" / BytesInteger(5, swapped=True),
)

TAIUTCDeltaStatus = Struct(
    *TAI_UTC_DeltaPaddedField("tai_utc_delta_current"),
    *TAI_UTC_DeltaPaddedField("tai_utc_delta_new"),
    "tai_of_delta_change" / BytesInteger(5, swapped=True),
)

TimeRoleGet = Struct()

TimeRoleSet = TimeRoleMsg

TimeRoleStatus = TimeRoleMsg


# fmt: off

class TimeOpcode(enum.IntEnum):
    TIME_GET = 0x8237
    TIME_SET = 0x005C
    TIME_STATUS = 0x005D
    TIME_ROLE_GET = 0x8238
    TIME_ROLE_SET = 0x8239
    TIME_ROLE_STATUS = 0x823A
    TIME_ZONE_GET = 0x823B
    TIME_ZONE_SET = 0x823C
    TIME_ZONE_STATUS = 0x823D
    TAI_UTC_DELTA_GET = 0x823E
    TAI_UTC_DELTA_SET = 0x823F
    TAI_UTC_DELTA_STATUS = 0x8240


# fmt: off
TimeMessage = SwitchStruct(
    "opcode" / Opcode(TimeOpcode),
    "params" / Switch(
        this.opcode,
        {
            TimeOpcode.TIME_GET: TimeGet,
            TimeOpcode.TIME_SET: TimeSet,
            TimeOpcode.TIME_STATUS: TimeStatus,
            TimeOpcode.TIME_ZONE_GET: TimeZoneGet,
            TimeOpcode.TIME_ZONE_SET: TimeZoneSet,
            TimeOpcode.TIME_ZONE_STATUS: TimeZoneStatus,
            TimeOpcode.TAI_UTC_DELTA_GET: TAIUTCDeltaGet,
            TimeOpcode.TAI_UTC_DELTA_SET: TAIUTCDeltaSet,
            TimeOpcode.TAI_UTC_DELTA_STATUS: TAIUTCDeltaStatus,
            TimeOpcode.TIME_ROLE_GET: TimeRoleGet,
            TimeOpcode.TIME_ROLE_SET: TimeRoleSet,
            TimeOpcode.TIME_ROLE_STATUS: TimeRoleStatus
        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Embedded, Int8ul, Int16sl, Int32sl, Struct, this

from bluetooth_mesh.messages.generics import OptionalSetParameters, TransitionTime, TransitionTimeAdapter
from bluetooth_mesh.messages.util import EnumSwitch as Switch, NamedSelect, Opcode, SwitchStruct


class GenericLevelOpcode(IntEnum):
    GENERIC_LEVEL_GET = 0x8205
    GENERIC_LEVEL_SET = 0x8206
    GENERIC_LEVEL_SET_UNACKNOWLEDGED = 0x8207
    GENERIC_LEVEL_STATUS = 0x8208
    GENERIC_DELTA_SET = 0x8209
    GENERIC_DELTA_SET_UNACKNOWLEDGED = 0x820A
    GENERIC_MOVE_SET = 0x820B
    GENERIC_MOVE_SET_UNACKNOWLEDGED = 0x820C


# fmt: off
GenericLevelGet = Struct()

GenericLevelSetMinimal = Struct(
    "level" / Int16sl,
    "tid" / Int8ul
)

GenericLevelSetOptional = Struct(
    Embedded(GenericLevelSetMinimal),
    Embedded(OptionalSetParameters)
)

GenericLevelSet = NamedSelect(
    optional=GenericLevelSetOptional,
    minimal=GenericLevelSetMinimal
)

GenericDeltaSetMinimal = Struct(
    "delta_level" / Int32sl,
    "tid" / Int8ul
)

GenericDeltaSetOptional = Struct(
    Embedded(GenericDeltaSetMinimal),
    Embedded(OptionalSetParameters)
)

GenericDeltaSet = NamedSelect(
    optional=GenericDeltaSetOptional,
    minimal=GenericDeltaSetMinimal
)

GenericMoveSetMinimal = Struct(
    "delta_level" / Int16sl,
    "tid" / Int8ul
)

GenericMoveSetOptional = Struct(
    Embedded(GenericMoveSetMinimal),
    Embedded(OptionalSetParameters)
)

GenericMoveSet = NamedSelect(
    optional=GenericMoveSetOptional,
    minimal=GenericMoveSetMinimal
)

GenericLevelStatusMinimal = Struct(
    "present_level" / Int16sl
)

GenericLevelStatusOptional = Struct(
    Embedded(GenericLevelStatusMinimal),
    "target_level" / Int16sl,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True)
)

GenericLevelStatus = NamedSelect(
    optional=GenericLevelStatusOptional,
    minimal=GenericLevelStatusMinimal
)

GenericLevelMessage = SwitchStruct(
    "opcode" / Opcode(GenericLevelOpcode),
    "params" / Switch(
        this.opcode,
        {
            GenericLevelOpcode.GENERIC_LEVEL_GET: GenericLevelGet,
            GenericLevelOpcode.GENERIC_LEVEL_SET: GenericLevelSet,
            GenericLevelOpcode.GENERIC_LEVEL_SET_UNACKNOWLEDGED: GenericLevelSet,
            GenericLevelOpcode.GENERIC_LEVEL_STATUS: GenericLevelStatus,
            GenericLevelOpcode.GENERIC_DELTA_SET: GenericDeltaSet,
            GenericLevelOpcode.GENERIC_DELTA_SET_UNACKNOWLEDGED: GenericDeltaSet,
            GenericLevelOpcode.GENERIC_MOVE_SET: GenericMoveSet,
            GenericLevelOpcode.GENERIC_MOVE_SET_UNACKNOWLEDGED: GenericMoveSet,
        },
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2022  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    BitsInteger,
    BitStruct,
    ByteSwapped,
    Embedded,
    GreedyBytes,
    GreedyRange,
    Int8sl,
    Int8ul,
    Int16sl,
    Int16ul,
    Int32ul,
    Prefixed,
    PrefixedArray,
    Struct,
    this,
)

from bluetooth_mesh.messages.util import (
    EnumAdapter,
    EnumSwitch as Switch,
    NamedSelect,
    Opcode,
    SetAdapter,
    SwitchStruct,
)


class RRuleSchedulerOpcode(IntEnum):
    SILVAIR_RRULE_SCHEDULER = 0xE83601


class RRuleSchedulerSubOpcode(IntEnum):
    RULES_LIST_GET = 0x00
    RULES_LIST_STATUS = 0x01
    REGISTER_MAX_SIZE_GET = 0x02
    REGISTER_MAX_SIZE_STATUS = 0x03
    SCHEDULER_MODE_GET = 0x04
    SCHEDULER_MODE_SET = 0x05
    SCHEDULER_MODE_STATUS = 0x06
    SCHEDULE_REGISTER_ENTRY_GET = 0x07
    SCHEDULE_REGISTER_ENTRY_SET = 0x08
    SCHEDULE_REGISTER_ENTRY_DELETE = 0x09
    SCHEDULE_REGISTER_ENTRY_DELETE_ALL = 0x0A
    SCHEDULE_REGISTER_ENTRY_STATUS = 0x0B
    SCHEDULE_REGISTER_LIST_GET = 0x0C
    SCHEDULE_REGISTER_LIST_STATUS = 0x0D


class SchedulerModes(IntEnum):
    DISABLED = 0x00
    ENABLED = 0x01


class Days(IntEnum):
    PREVIOUS_SUNDAY = -7
    PREVIOUS_SATURDAY = -6
    PREVIOUS_FRIDAY = -5
    PREVIOUS_THURSDAY = -4
    PREVIOUS_WEDNESDAY = -3
    PREVIOUS_TUESDAY = -2
    PREVIOUS_MONDAY = -1
    NEXT_MONDAY = 1
    NEXT_TUESDAY = 2
    NEXT_WEDNESDAY = 3
    NEXT_THURSDAY = 4
    NEXT_FRIDAY = 5
    NEXT_SATURDAY = 6
    NEXT_SUNDAY = 7


class StatusCodes(IntEnum):
    SUCCESS = 0x00
    INVALID_SLOT = 0x01
    INVALID_SLOT_PARAMETERS = 0x02
    INVALID_RULE = 0x03
    INVALID_RULE_VALUE = 0x04
    REGISTER_SIZE_EXCEEDED = 0x05
    STORAGE_FAILURE = 0x06


class RuleIDs(IntEnum):
    FREQ = 0x00
    UNTIL = 0x01
    COUNT = 0x02
    INTERVAL = 0x03
    BYSECOND = 0x04
    BYMINUTE = 0x05
    BYHOUR = 0x06
    BYDAY = 0x07
    BYMONTHDAY = 0x08
    BYYEARDAY = 0x09
    BYWEEKNO = 0x0A
    BYMONTH = 0x0B
    BYSETPOS = 0x0C
    DTSTART = 0x0D
    EXPLICIT_RDATE = 0x0E
    EXCLUSIONS_RDATE = 0x0F


class Freqs(IntEnum):
    SECONDLY = 0x00
    MINUTELY = 0x01
    HOURLY = 0x02
    DAILY = 0x03
    WEEKLY = 0x04
    MONTHLY = 0x05
    YEARLY = 0x06


# fmt: off
Slots = Struct(
    "slot_id" / Int16ul,
    "element" / Int8ul,
    "slot_parameter" / Prefixed(Int8ul, GreedyBytes),
)

DateTime = ByteSwapped(
    BitStruct(
        "second" / BitsInteger(6),
        "minute" / BitsInteger(6),
        "hour" / BitsInteger(5),
        "day" / BitsInteger(5),
        "month" / BitsInteger(4),
        "year" / BitsInteger(14),
    )
)

RuleID = EnumAdapter(Int8ul, RuleIDs)

Freq = EnumAdapter(Int8ul, Freqs)

Day = EnumAdapter(Int8sl, Days)

Rules = Struct(
    "rule_id" / RuleID,
    "rule" / Switch(
        this.rule_id,
        {
            RuleIDs.FREQ: Freq,
            RuleIDs.UNTIL: DateTime,
            RuleIDs.COUNT: Int16ul,
            RuleIDs.INTERVAL: Int16ul,
            RuleIDs.BYSECOND: PrefixedArray(Int8ul, Int8ul),
            RuleIDs.BYMINUTE: PrefixedArray(Int8ul, Int8ul),
            RuleIDs.BYHOUR: PrefixedArray(Int8ul, Int8ul),
            RuleIDs.BYDAY: PrefixedArray(Int8ul, Day),
            RuleIDs.BYMONTHDAY: PrefixedArray(Int8ul, Int8sl),
            RuleIDs.BYYEARDAY: PrefixedArray(Int8ul, Int16sl),
            RuleIDs.BYWEEKNO: PrefixedArray(Int8ul, Int8sl),
            RuleIDs.BYMONTH: Int16ul,
            RuleIDs.BYSETPOS: PrefixedArray(Int8ul, Int16sl),
            RuleIDs.DTSTART: DateTime,
            RuleIDs.EXPLICIT_RDATE: PrefixedArray(Int8ul, DateTime),
            RuleIDs.EXCLUSIONS_RDATE: PrefixedArray(Int8ul, DateTime),
        },
    ),
)

SchedulerRegisterEntry = Struct(
    "array_of_slots" / PrefixedArray(Int8ul, Slots),
    "rules" / GreedyRange(Rules)
)

RulesListGet = Struct()

RulesListStatus = Struct(
    "rule_ids" / SetAdapter(GreedyRange(RuleID))
)

RegisterMaxSizeGet = Struct()

RegisterMaxSizeStatus = Struct(
    "register_max_size" / Int32ul
)

SchedulerModeGet = Struct()

SchedulerModeSet = Struct(
    "scheduler_mode" / EnumAdapter(Int8ul, SchedulerModes)
)

SchedulerModeStatus = SchedulerModeSet

ScheduleRegisterEntryGet = Struct(
    "entry_id" / Int8ul
)

ScheduleRegisterEntrySet = Struct(
    "entry_id" / Int8ul,
    "scheduler_register_entry" / SchedulerRegisterEntry
)

ScheduleRegisterEntryDelete = ScheduleRegisterEntryGet

ScheduleRegisterEntryDeleteAll = Struct()

ScheduleRegisterEntryStatusMinimal = Struct(
    "entry_id" / Int8ul,
    "status" / EnumAdapter(Int8ul, StatusCodes),
)

ScheduleRegisterEntryStatusOptional = Struct(
    Embedded(ScheduleRegisterEntryStatusMinimal),
    "scheduler_register_entry" / SchedulerRegisterEntry,
)

ScheduleRegisterEntryStatus = NamedSelect(
    optional=ScheduleRegisterEntryStatusOptional,
    minimal=ScheduleRegisterEntryStatusMinimal
)

ScheduleRegisterListGet = Struct()

ScheduleRegisterListStatus = Struct(
    "entry_ids" / SetAdapter(GreedyRange(Int8ul))
)

RRuleSchedulerSubMessage = Struct(
    "subopcode" / EnumAdapter(Int8ul, RRuleSchedulerSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            RRuleSchedulerSubOpcode.RULES_LIST_GET: RulesListGet,
            RRuleSchedulerSubOpcode.RULES_LIST_STATUS: RulesListStatus,
            RRuleSchedulerSubOpcode.REGISTER_MAX_SIZE_GET: RegisterMaxSizeGet,
            RRuleSchedulerSubOpcode.REGISTER_MAX_SIZE_STATUS: RegisterMaxSizeStatus,
            RRuleSchedulerSubOpcode.SCHEDULER_MODE_GET: SchedulerModeGet,
            RRuleSchedulerSubOpcode.SCHEDULER_MODE_SET: SchedulerModeSet,
            RRuleSchedulerSubOpcode.SCHEDULER_MODE_STATUS: SchedulerModeStatus,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_GET: ScheduleRegisterEntryGet,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_SET: ScheduleRegisterEntrySet,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_DELETE: ScheduleRegisterEntryDelete,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_DELETE_ALL: ScheduleRegisterEntryDeleteAll,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_STATUS: ScheduleRegisterEntryStatus,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_LIST_GET: ScheduleRegisterListGet,
            RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_LIST_STATUS: ScheduleRegisterListStatus,
        },
    ),
)

RRuleSchedulerMessage = SwitchStruct(
    "opcode" / Opcode(RRuleSchedulerOpcode),
    "params" / Switch(
        this.opcode,
        {
            RRuleSchedulerOpcode.SILVAIR_RRULE_SCHEDULER: RRuleSchedulerSubMessage,
        },
    ),
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages import AccessMessage

valid = [
    # fmt: off
    pytest.param(
        bytes.fromhex("82310200"),
        {
            "opcode": "SENSOR_GET",
            "params": {
                "property_id": "AVERAGE_INPUT_CURRENT",
            }
        },
        id="SensorGet[AVERAGE_INPUT_CURRENT]"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,decoded", valid)
def test_build(encoded, decoded):
    result = AccessMessage.build(obj=decoded)
    assert result == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import datetime, timedelta, timezone

import pytest

from bluetooth_mesh.messages.time import (
    MESH_UNIX_EPOCH_DIFF,
    TimeMessage,
    TimeOpcode,
    TimeRole,
    mesh_tai_utc_delta_to_timedelta,
    mesh_time_zone_offset_to_timedelta,
    subsecond_to_seconds,
)

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x37',
        TimeOpcode.TIME_GET,
        dict(),
        id="TIME_GET"
    ),
    pytest.param(
        b'\x5C\xbd\xf7\x67\x26\x00\x1a\xb2\x49\x02\x48',
        TimeOpcode.TIME_SET,
        dict(
            date=datetime.fromtimestamp(
                0x2667f7bd + MESH_UNIX_EPOCH_DIFF + subsecond_to_seconds(0x1A) + mesh_time_zone_offset_to_timedelta(
                    0x48).total_seconds(),
                timezone(mesh_time_zone_offset_to_timedelta(0x48))),
            uncertainty=timedelta(milliseconds=(0xB2 * 10)),
            tai_utc_delta=mesh_tai_utc_delta_to_timedelta(0x0248 >> 1),
            time_authority=True,
        ),
        id="TIME_SET"
    ),
    pytest.param(
        b'\x5D\xbd\xf7\x67\x26\x00\x1a\xb2\x49\x02\x48',
        TimeOpcode.TIME_STATUS,
        dict(
            date=datetime.fromtimestamp(
                0x2667f7bd + MESH_UNIX_EPOCH_DIFF + subsecond_to_seconds(0x1A) + mesh_time_zone_offset_to_timedelta(
                    0x48).total_seconds(),
                timezone(mesh_time_zone_offset_to_timedelta(0x48))),
            uncertainty=timedelta(milliseconds=(0xB2 * 10)),
            tai_utc_delta=mesh_tai_utc_delta_to_timedelta(0x0248 >> 1),
            time_authority=True,
        ),
        id="TIME_STATUS"
    ),
    pytest.param(
        bytes.fromhex("823b"),
        TimeOpcode.TIME_ZONE_GET,
        dict(),
        id="TIME_ZONE_GET"
    ),
    pytest.param(
        bytes.fromhex("823cab3400000012"),
        TimeOpcode.TIME_ZONE_SET,
        dict(
            time_zone_offset_new=0xab,
            tai_of_zone_change=0x1200000034
        ),
        id="TIME_ZONE_SET"
    ),
    pytest.param(
        bytes.fromhex("823dcdab3400000012"),
        TimeOpcode.TIME_ZONE_STATUS,
        dict(
            time_zone_offset_current=0xcd,
            time_zone_offset_new=0xab,
            tai_of_zone_change=0x1200000034
        ),
        id="TIME_ZONE_STATUS"
    ),
    pytest.param(
        bytes.fromhex("823e"),
        TimeOpcode.TAI_UTC_DELTA_GET,
        dict(),
        id="TAI_UTC_DELTA_GET"
    ),
    pytest.param(
        bytes.fromhex("823f01005544332211"),
        TimeOpcode.TAI_UTC_DELTA_SET,
        dict(
            tai_utc_delta_new=0b000000000000001,
            tai_of_delta_change=0x1122334455
        ),
        id="TAI_UTC_DELTA_SET"
    ),
    pytest.param(
        bytes.fromhex("8240014001005544332211"),
        TimeOpcode.TAI_UTC_DELTA_STATUS,
        dict(
            tai_utc_delta_current=0b100000000000001,
            tai_utc_delta_new=0b000000000000001,
            tai_of_delta_change=0x1122334455
        ),
        id="TAI_UTC_DELTA_STATUS"
    ),
    pytest.param(
        bytes.fromhex("8238"),
        TimeOpcode.TIME_ROLE_GET,
        dict(),
        id="TIME_ROLE_GET"
    ),
    pytest.param(
        bytes.fromhex("823903"),
        TimeOpcode.TIME_ROLE_SET,
        dict(
            time_role=TimeRole.TIME_CLIENT,
        ),
        id="TIME_ROLE_SET"
    ),
    pytest.param(
        bytes.fromhex("823A03"),
        TimeOpcode.TIME_ROLE_STATUS,
        dict(
            time_role=TimeRole.TIME_CLIENT,
        ),
        id="TIME_ROLE_STATUS"
    ),
    # # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert TimeMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert TimeMessage.build(dict(opcode=opcode, params=data)) == encoded



import pytest

from bluetooth_mesh.messages.silvair.rrule_scheduler import (
    Days,
    Freqs,
    RRuleSchedulerSubMessage,
    RRuleSchedulerSubOpcode,
    RuleIDs,
    SchedulerModes,
    StatusCodes,
)

valid = [
    pytest.param(
        bytes.fromhex("080501CDAB00000001030200020500"),
        RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_SET,
        {
            "entry_id": 0x05,
            "slots": [{"slot_id": 0xABCD, "element": 0x00, "slot_parameter": b""}],
            "rules": [
                {"rule_id": RuleIDs.FREQ, "rule": Freqs.MINUTELY},
                {"rule_id": RuleIDs.INTERVAL, "rule": 0x02},
                {"rule_id": RuleIDs.COUNT, "rule": 0x05},
            ],
        },
    ),
    pytest.param(
        bytes.fromhex(
            "080503CDAB0002BBCCCCAB0205ABFF157288CBAB05100002BB" "CCCCAB0005ABFF157288CBAB000001030200020500"
        ),
        RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_SET,
        {
            "entry_id": 0x05,
            "slots": [
                {
                    "slot_id": 0xABCD,
                    "element": 0x00,
                    "slot_parameter": bytes.fromhex("BBCC"),
                },
                {
                    "slot_id": 0xABCC,
                    "element": 0x02,
                    "slot_parameter": bytes.fromhex("ABFF157288"),
                },
                {
                    "slot_id": 0xABCB,
                    "element": 0x05,
                    "slot_parameter": bytes.fromhex("0002BBCCCCAB0005ABFF157288CBAB00"),
                },
            ],
            "rules": [
                {"rule_id": RuleIDs.FREQ, "rule": Freqs.MINUTELY},
                {"rule_id": RuleIDs.INTERVAL, "rule": 0x02},
                {"rule_id": RuleIDs.COUNT, "rule": 0x05},
            ],
        },
    ),
    pytest.param(
        bytes.fromhex(
            "080503CDAB0002BBCCCCAB0205ABFF157288CBAB05100002BB"
            "CCCCAB0005ABFF157288CBAB00000103201002055004010005"
            "02010B0603020C100704010507FA0805041011121309080400"
            "10000200300040005000000120010A040616262A0B01020C06"
            "0A001000A000D000500050010DF8C7661BE101F8C7661BE10E"
            "03E4C7661BE1EEC7661BE1F8C7661BE10F03E4C7661BE1EEC7"
            "661BE1F8C7661BE1"
        ),
        RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_SET,
        {
            "entry_id": 0x05,
            "slots": [
                {
                    "slot_id": 0xABCD,
                    "element": 0x00,
                    "slot_parameter": bytes.fromhex("BBCC"),
                },
                {
                    "slot_id": 0xABCC,
                    "element": 0x02,
                    "slot_parameter": bytes.fromhex("ABFF157288"),
                },
                {
                    "slot_id": 0xABCB,
                    "element": 0x05,
                    "slot_parameter": bytes.fromhex("0002BBCCCCAB0005ABFF157288CBAB00"),
                },
            ],
            "rules": [
                {"rule_id": RuleIDs.FREQ, "rule": Freqs.MINUTELY},
                {"rule_id": RuleIDs.INTERVAL, "rule": 0x1020},
                {"rule_id": RuleIDs.COUNT, "rule": 0x5005},
                {"rule_id": RuleIDs.BYSECOND, "rule": [0x00]},
                {"rule_id": RuleIDs.BYMINUTE, "rule": [0x01, 0x0B]},
                {"rule_id": RuleIDs.BYHOUR, "rule": [0x02, 0x0C, 0x10]},
                {
                    "rule_id": RuleIDs.BYDAY,
                    "rule": [
                        Days.NEXT_MONDAY,
                        Days.NEXT_FRIDAY,
                        Days.NEXT_SUNDAY,
                        Days.PREVIOUS_SATURDAY,
                    ],
                },
                {"rule_id": RuleIDs.BYMONTHDAY, "rule": [0x04, 0x10, 0x11, 0x12, 0x13]},
                {
                    "rule_id": RuleIDs.BYYEARDAY,
                    "rule": [
                        0x0004,
                        0x0010,
                        0x0002,
                        0x0030,
                        0x0040,
                        0x0050,
                        0x0100,
                        0x0120,
                    ],
                },
                {"rule_id": RuleIDs.BYWEEKNO, "rule": [0x06, 0x16, 0x26, 0x2A]},
                {"rule_id": RuleIDs.BYMONTH, "rule": 0x0201},
                {
                    "rule_id": RuleIDs.BYSETPOS,
                    "rule": [0x000A, 0x0010, 0x00A0, 0x00D0, 0x0050, 0x0150],
                },
                {
                    "rule_id": RuleIDs.DTSTART,
                    "rule": dict(second=56, minute=17, hour=22, day=25, month=11, year=2040),
                },
                {
                    "rule_id": RuleIDs.UNTIL,
                    "rule": dict(second=56, minute=17, hour=22, day=25, month=11, year=2040),
                },
                {
                    "rule_id": RuleIDs.EXPLICIT_RDATE,
                    "rule": [
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2020),
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2030),
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2040),
                    ],
                },
                {
                    "rule_id": RuleIDs.EXCLUSIONS_RDATE,
                    "rule": [
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2020),
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2030),
                        dict(second=56, minute=17, hour=22, day=25, month=11, year=2040),
                    ],
                },
            ],
        },
    ),
]


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_parse_schedule_register_entry_format(encoded, subopcode, payload):
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == subopcode

    assert decoded.payload.entry_id == (payload["entry_id"])

    for idx, sl in enumerate(decoded.payload.scheduler_register_entry.array_of_slots):
        assert sl["slot_id"] == payload["slots"][idx]["slot_id"]
        assert sl["element"] == payload["slots"][idx]["element"]
        assert sl["slot_parameter"] == payload["slots"][idx]["slot_parameter"]

    for idx, rl in enumerate(decoded.payload.scheduler_register_entry.rules):
        assert rl["rule_id"] == payload["rules"][idx]["rule_id"]
        assert rl["rule"] == payload["rules"][idx]["rule"]


def test_scheduler_rules_list_status():
    encoded = bytes.fromhex("010102030405")
    payload = {
        RuleIDs.BYMINUTE,
        RuleIDs.BYSECOND,
        RuleIDs.INTERVAL,
        RuleIDs.UNTIL,
        RuleIDs.COUNT,
    }
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == RRuleSchedulerSubOpcode.RULES_LIST_STATUS
    assert decoded.payload.rule_ids == payload


def test_scheduler_max_size_status():
    encoded = bytes.fromhex("03EFCDAB89")
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == RRuleSchedulerSubOpcode.REGISTER_MAX_SIZE_STATUS
    assert decoded.payload.register_max_size == 0x89ABCDEF


def test_scheduler_mode_set():
    encoded = bytes.fromhex("0501")
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == RRuleSchedulerSubOpcode.SCHEDULER_MODE_SET
    assert decoded.payload.scheduler_mode == SchedulerModes.ENABLED


def test_scheduler_register_entry_status_no_register_entry():
    encoded = bytes.fromhex("0BFE03")
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_STATUS
    assert decoded.payload.entry_id == 0xFE
    assert decoded.payload.status == StatusCodes.INVALID_RULE


def test_scheduler_register_entry_status_with_register_entry():
    encoded = bytes.fromhex("0BFE0001CDAB00020A0B000203020002050006020F17")
    decoded = RRuleSchedulerSubMessage.parse(encoded)
    assert decoded.subopcode == RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_ENTRY_STATUS
    assert decoded.payload.entry_id == 0xFE
    assert decoded.payload.status == StatusCodes.SUCCESS

    print(decoded)
    register_entry = {
        "entry_id": 0x05,
        "slots": [
            {
                "slot_id": 0xABCD,
                "element": 0x00,
                "slot_parameter": bytes.fromhex("0A0B"),
            }
        ],
        "rules": [
            {"rule_id": RuleIDs.FREQ, "rule": Freqs.HOURLY},
            {"rule_id": RuleIDs.INTERVAL, "rule": 0x02},
            {"rule_id": RuleIDs.COUNT, "rule": 0x05},
            {"rule_id": RuleIDs.BYHOUR, "rule": [15, 23]},
        ],
    }
    for idx, sl in enumerate(decoded.payload.scheduler_register_entry.array_of_slots):
        assert sl["slot_id"] == register_entry["slots"][idx]["slot_id"]
        assert sl["element"] == register_entry["slots"][idx]["element"]
        assert sl["slot_parameter"] == register_entry["slots"][idx]["slot_parameter"]

    for idx, rl in enumerate(decoded.payload.scheduler_register_entry.rules):
        assert rl["rule_id"] == register_entry["rules"][idx]["rule_id"]
        assert rl["rule"] == register_entry["rules"][idx]["rule"]


def test_scheduler_register_list_status():
    encoded = bytes.fromhex("0D01050A")
    decoded = RRuleSchedulerSubMessage.parse(encoded)

    assert decoded.subopcode == RRuleSchedulerSubOpcode.SCHEDULE_REGISTER_LIST_STATUS
    assert decoded.payload.entry_ids == {1, 5, 10}



from datetime import datetime, timedelta, timezone

import pytest

from bluetooth_mesh.messages.properties import PropertyID
from bluetooth_mesh.messages.silvair.emergency_lighting_test import (
    EmergencyLightingTestServerParams,
    EmergencyLightingTestServerSubOpcode,
    TestExecutionStatus,
)

valid = [
    pytest.param(
        bytes.fromhex("00"),
        EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("01"),
        EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_START,
        {},
    ),
    pytest.param(
        bytes.fromhex("02"),
        EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_STOP,
        {},
    ),
    pytest.param(
        bytes.fromhex("0305040302010608070506"),
        EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2137,
                    3,
                    2,
                    17,
                    2,
                    45,
                    # time_zone_offset is -870min, which is -52200s, but
                    # python encodes that as -1d +34200s
                    tzinfo=timezone(timedelta(days=-1, seconds=34200)),
                ),
                tai_utc_delta=timedelta(seconds=1800),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.DROPPED,
        ),
    ),
    pytest.param(
        bytes.fromhex("030000000000060807050604030201"),
        EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    1999,
                    12,
                    30,
                    19,
                    0,
                    tzinfo=timezone(timedelta(days=-1, seconds=34200)),
                ),
                tai_utc_delta=timedelta(seconds=1800),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.DROPPED,
            relative_timestamp=0x01020304,
        ),
    ),
    pytest.param(
        bytes.fromhex("04"),
        EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_GET,
        {},
    ),
    pytest.param(
        bytes.fromhex("05"),
        EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_START,
        {},
    ),
    pytest.param(
        bytes.fromhex("06"),
        EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_STOP,
        {},
    ),
    pytest.param(
        bytes.fromhex("07050403020106080705063412"),
        EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    2137,
                    3,
                    2,
                    17,
                    2,
                    45,
                    tzinfo=timezone(timedelta(days=-1, seconds=34200)),
                ),
                tai_utc_delta=timedelta(seconds=1800),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.DROPPED,
            duration_result=0x1234,
        ),
    ),
    pytest.param(
        bytes.fromhex("0700000000000608070506341204030201"),
        EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_STATUS,
        dict(
            tai_timestamp=dict(
                date=datetime(
                    1999,
                    12,
                    30,
                    19,
                    0,
                    tzinfo=timezone(timedelta(days=-1, seconds=34200)),
                ),
                tai_utc_delta=timedelta(seconds=1800),
            ),
            execution_result=dict(
                lamp_fault=False,
                battery_fault=True,
                circuit_fault=True,
                battery_duration_fault=False,
            ),
            execution_status=TestExecutionStatus.DROPPED,
            duration_result=0x1234,
            relative_timestamp=0x01020304,
        ),
    ),
    pytest.param(
        bytes.fromhex("0886FF"),
        EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_GET,
        dict(
            property_id=PropertyID.ELT_DURATION_TEST_RETRY_PERIOD,
        ),
    ),
    pytest.param(
        bytes.fromhex("0986FF04030201"),
        EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_SET,
        dict(
            property_id=PropertyID.ELT_DURATION_TEST_RETRY_PERIOD,
            value=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0A86FF04030201"),
        EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED,
        dict(
            property_id=PropertyID.ELT_DURATION_TEST_RETRY_PERIOD,
            value=dict(seconds=0x01020304),
        ),
    ),
    pytest.param(
        bytes.fromhex("0B86FF04030201"),
        EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_STATUS,
        dict(
            property_id=PropertyID.ELT_DURATION_TEST_RETRY_PERIOD,
            value=dict(seconds=0x01020304),
        ),
    ),
]


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_parse(encoded, subopcode, payload):
    decoded = EmergencyLightingTestServerParams.parse(encoded)
    assert decoded.subopcode == subopcode
    assert decoded.payload == payload


@pytest.mark.parametrize("encoded, subopcode, payload", valid)
def test_build(encoded, subopcode, payload):
    assert EmergencyLightingTestServerParams.build(dict(subopcode=subopcode, payload=payload)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Int8ul, Struct, this

from bluetooth_mesh.messages.generics import Delay, TransitionTime, TransitionTimeAdapter
from bluetooth_mesh.messages.util import EnumSwitch as Switch, NamedSelect, Opcode, SwitchStruct


class GenericOnOffOpcode(IntEnum):
    GENERIC_ONOFF_GET = 0x8201
    GENERIC_ONOFF_SET = 0x8202
    GENERIC_ONOFF_SET_UNACKNOWLEDGED = 0x8203
    GENERIC_ONOFF_STATUS = 0x8204


# fmt: off
GenericOnOffGet = Struct()

GenericOnOffSetMinimal = Struct(
    "onoff" / Int8ul,
    "tid" / Int8ul
)

GenericOnOffSetOptional = Struct(
    "onoff" / Int8ul,
    "tid" / Int8ul,
    "transition_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=False),
    "delay" / Delay(Int8ul),
)

GenericOnOffSet = NamedSelect(
    optional=GenericOnOffSetOptional,
    minimal=GenericOnOffSetMinimal
)

GenericOnOffStatusMinimal = Struct(
    "present_onoff" / Int8ul
)

GenericOnOffStatusOptional = Struct(
    "present_onoff" / Int8ul,
    "target_onoff" / Int8ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True),
)

GenericOnOffStatus = NamedSelect(
    optional=GenericOnOffStatusOptional,
    minimal=GenericOnOffStatusMinimal
)


GenericOnOffMessage = SwitchStruct(
    "opcode" / Opcode(GenericOnOffOpcode),
    "params" / Switch(
        this.opcode,
        {
            GenericOnOffOpcode.GENERIC_ONOFF_GET: GenericOnOffGet,
            GenericOnOffOpcode.GENERIC_ONOFF_SET: GenericOnOffSet,
            GenericOnOffOpcode.GENERIC_ONOFF_SET_UNACKNOWLEDGED: GenericOnOffSet,
            GenericOnOffOpcode.GENERIC_ONOFF_STATUS: GenericOnOffStatus,
        },
    ),
)
# fmt: off



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages import AccessMessage

valid = [
    # fmt: off
    pytest.param(
        bytes.fromhex("04003601030405"),
        {
            "opcode": 0x04,
            "params": {
                "test_id": 0,
                "company_id": 0x0136,
                "fault_array": [0x03, 0x04, 0x05]
            },
            "health_current_status": {
                "test_id": 0,
                "company_id": 0x0136,
                "fault_array": [0x03, 0x04, 0x05]
            }
        },
        id="1 byte opcode"
    ),
    pytest.param(
        bytes.fromhex("8032013601"),
        {
            "opcode": 0x8032,
            "params": {
                "test_id": 1,
                "company_id": 0x0136
            },
            "health_fault_test": {
                "test_id": 1,
                "company_id": 0x0136
            }
        },
        id="2 byte opcode"
    ),
    pytest.param(
        bytes.fromhex("c0112233"),
        {
            "opcode": 0xc01122,
            "params": bytes([0x33])
        },
        id="3 byte opcode"
    )
    # fmt: on
]


@pytest.mark.parametrize("encoded,decoded", valid)
def test_build(encoded, decoded):
    result = AccessMessage.build(obj=decoded)
    # print(result)
    assert result == encoded


@pytest.mark.parametrize("encoded,decoded", valid)
def test_parse(encoded, decoded):
    result = AccessMessage.parse(data=encoded)
    # print(result)
    assert result == decoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.generic.level import GenericLevelMessage, GenericLevelOpcode

valid = [
    # fmt: off
    pytest.param(
        b'\x82\x05',
        GenericLevelOpcode.GENERIC_LEVEL_GET,
        dict(),
        id="LEVEL_GET"
    ),
    pytest.param(
        b'\x82\x06\xff\x7f\x22',
        GenericLevelOpcode.GENERIC_LEVEL_SET,
        dict(
            level=32767,
            tid=34
        ),
        id="LEVEL_SET_max"
    ),
    pytest.param(
        b'\x82\x06\x00\x80\x22',
        GenericLevelOpcode.GENERIC_LEVEL_SET,
        dict(
            level=-32768,
            tid=34
        ),
        id="LEVEL_SET_min"
    ),
    pytest.param(
        b'\x82\x06\x01\x00\x22',
        GenericLevelOpcode.GENERIC_LEVEL_SET,
        dict(
            level=1,
            tid=34,
            transition_time=6.3,
            delay=0.3
        ),
        id="LEVEL_SET_invalid"
    ),
    pytest.param(
        b'\x82\x06\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_LEVEL_SET,
        dict(
            level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="LEVEL_SET_with_optional"
    ),
    pytest.param(
        b'\x82\x07\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_LEVEL_SET_UNACKNOWLEDGED,
        dict(
            level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="LEVEL_SET_UNACKNOWLEDGED"
    ),
    pytest.param(
        b'\x82\x08\xff\x7f',
        GenericLevelOpcode.GENERIC_LEVEL_STATUS,
        dict(
            present_level=32767,
            target_level=None,
            remaining_time=None
        ),
        id="LEVEL_STATUS_max"
    ),
    pytest.param(
        b'\x82\x08\x00\x80',
        GenericLevelOpcode.GENERIC_LEVEL_STATUS,
        dict(
            present_level=-32768,
            target_level=None,
            remaining_time=None
        ),
        id="LEVEL_STATUS_min"
    ),
    pytest.param(
        b'\x82\x08\x00\x00\xff\x00\x4a',
        GenericLevelOpcode.GENERIC_LEVEL_STATUS,
        dict(
            present_level=0,
            target_level=255,
            remaining_time=10
        ),
        id="LEVEL_STATUS_with_optional"
    ),
    pytest.param(
        b'\x82\x08\x00\x00\x01\x00\xff',
        GenericLevelOpcode.GENERIC_LEVEL_STATUS,
        dict(
            present_level=0,
            target_level=1,
            remaining_time=37800
        ),
        id="LEVEL_STATUS_max_time"
    ),
    pytest.param(
        b'\x82\x09\xff\xff\xff\x7f\x22',
        GenericLevelOpcode.GENERIC_DELTA_SET,
        dict(
            delta_level=2147483647,
            tid=34
        ),
        id="DELTA_SET_max"
    ),
    pytest.param(
        b'\x82\x09\x00\x00\x00\x80\x22',
        GenericLevelOpcode.GENERIC_DELTA_SET,
        dict(
            delta_level=-2147483648,
            tid=34
        ),
        id="DELTA_SET_min"
    ),
    pytest.param(
        b'\x82\x09\x01\x00\x00\x00\x22',
        GenericLevelOpcode.GENERIC_DELTA_SET,
        dict(
            delta_level=1,
            tid=34,
            transition_time=6.3,
            delay=0.3
        ),
        id="DELTA_SET_invalid"),
    pytest.param(
        b'\x82\x09\x00\x00\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_DELTA_SET,
        dict(
            delta_level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="DELTA_SET_with_optional"),
    pytest.param(
        b'\x82\x0a\x00\x00\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_DELTA_SET_UNACKNOWLEDGED,
        dict(
            delta_level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="DELTA_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x0b\xff\x7f\x22',
        GenericLevelOpcode.GENERIC_MOVE_SET,
        dict(
            delta_level=32767,
            tid=34
        ),
        id="MOVE_SET_max"
    ),
    pytest.param(
        b'\x82\x0b\x00\x80\x22',
        GenericLevelOpcode.GENERIC_MOVE_SET,
        dict(
            delta_level=-32768,
            tid=34
        ),
        id="MOVE_SET_min"
    ),
    pytest.param(
        b'\x82\x0b\x01\x00\x22',
        GenericLevelOpcode.GENERIC_MOVE_SET,
        dict(
            delta_level=1,
            tid=34,
            transition_time=6.3,
            delay=0.3
        ),
        id="MOVE_SET_invalid"
    ),
    pytest.param(
        b'\x82\x0b\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_MOVE_SET,
        dict(
            delta_level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="MOVE_SET_with_optional"
    ),
    pytest.param(
        b'\x82\x0c\x00\x00\x31\x32\x3c',
        GenericLevelOpcode.GENERIC_MOVE_SET_UNACKNOWLEDGED,
        dict(
            delta_level=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="MOVE_SET_UNACKNOWLEDGED"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert GenericLevelMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert GenericLevelMessage.build(dict(opcode=opcode, params=data)) == encoded



from .debug import DebugMessage, DebugSubOpcode
from .debugV2 import DebugV2Message, DebugV2Opcode
from .emergency_lighting_test import EmergencyLightingTestServerMessage, EmergencyLightingTestServerSubOpcode
from .gateway_config_server import GatewayConfigMessage, GatewayConfigServerSubOpcode
from .light_extended_controller import LightExtendedControllerMessage, LightExtendedControllerSubOpcode
from .network_diagnostic_server import (
    NetworkDiagnosticServerMessage,
    NetworkDiagnosticServerSubOpcode,
    NetworkDiagnosticSetupServerMessage,
    NetworkDiagnosticSetupServerSubOpcode,
)
from .rrule_scheduler import RRuleSchedulerMessage, RRuleSchedulerOpcode



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import io
from enum import IntEnum

from construct import (
    BitsInteger,
    Construct,
    Embedded,
    GreedyRange,
    Int8ul,
    Int16ul,
    Select,
    Struct,
    Switch,
    stream_read,
    stream_read_entire,
    stream_write,
    this,
)

from bluetooth_mesh.messages.config import DoubleKeyIndex, EmbeddedBitStruct
from bluetooth_mesh.messages.properties import DefaultCountValidator, PropertyDict, PropertyID, PropertyValue
from bluetooth_mesh.messages.util import (
    AliasedContainer,
    EnumAdapter,
    FieldAdapter,
    NamedSelect,
    Opcode,
    SwitchStruct,
)


class SensorSampling(IntEnum):
    UNSPECIFIED = 0x00
    INSTANTANEOUS = 0x01
    ARITHMETIC_MEAN = 0x02
    RMS = 0x03
    MAXIMUM = 0x04
    MINIMUM = 0x05
    ACCUMULATED = 0x06
    COUNT = 0x07


class SensorSettingAccess(IntEnum):
    READ_ONLY = 0x01
    READ_WRITE = 0x03


class SensorOpcode(IntEnum):
    SENSOR_DESCRIPTOR_GET = 0x8230
    SENSOR_DESCRIPTOR_STATUS = 0x51
    SENSOR_GET = 0x8231
    SENSOR_STATUS = 0x52
    SENSOR_COLUMN_GET = 0x8232
    SENSOR_COLUMN_STATUS = 0x53
    SENSOR_SERIES_GET = 0x8233
    SENSOR_SERIES_STATUS = 0x54

    def __repr__(self):
        return str(self.value)


class SensorSetupOpcode(IntEnum):
    SENSOR_CADENCE_GET = 0x8234
    SENSOR_CADENCE_SET = 0x55
    SENSOR_CADENCE_SET_UNACKNOWLEDGED = 0x56
    SENSOR_CADENCE_STATUS = 0x57
    SENSOR_SETTINGS_GET = 0x8235
    SENSOR_SETTINGS_STATUS = 0x58
    SENSOR_SETTING_GET = 0x8236
    SENSOR_SETTING_SET = 0x59
    SENSOR_SETTING_SET_UNACKNOWLEDGED = 0x5A
    SENSOR_SETTING_STATUS = 0x5B

    def __repr__(self):
        return str(self.value)


# fmt: off
SensorPropertyId = FieldAdapter(
    Select(
        EnumAdapter(Int16ul, PropertyID),
        Int16ul),
    Int16ul
)

SensorGetMinimal = Struct()

SensorGetOptional = Struct(
    "property_id" / SensorPropertyId,
)

SensorGet = NamedSelect(
    optional=SensorGetOptional,
    minimal=SensorGetMinimal
)

SensorSettingsGet = Struct(
    "sensor_property_id" / SensorPropertyId,
)

SensorSettingGet = Struct(
    Embedded(SensorSettingsGet),
    "sensor_setting_property_id" / SensorPropertyId,
)

class SensorSettingRawMixin:
    def _parse_sensor_setting(self, stream, context, path, sensor_setting_property_id, **kwargs):
        try:
            sensor_setting_property_id = PropertyID(sensor_setting_property_id)
            sensor_setting_name = sensor_setting_property_id.name.lower()
        except ValueError:
            sensor_setting_name = "sensor_setting_raw"

        try:
            sensor_setting_raw = PropertyDict[sensor_setting_property_id]._parse(stream, context, path)
        except KeyError:
            sensor_setting_raw = list(stream_read_entire(stream))

        class _Container(AliasedContainer):
            ALIAS = sensor_setting_name
            ORIGINAL = "sensor_setting_raw"

        return _Container({
            **kwargs,
            "sensor_setting_property_id": sensor_setting_property_id,
            sensor_setting_name: sensor_setting_raw
        })

    def _build_sensor_setting(self, obj, stream, context, path, sensor_setting_property_id):
        try:
            sensor_setting_property_id = PropertyID(sensor_setting_property_id)
            sensor_setting_name = sensor_setting_property_id.name.lower()
        except ValueError:
            sensor_setting_name = "sensor_setting_raw"

        sensor_setting_raw = obj.get(sensor_setting_name, obj.get("sensor_setting_raw"))

        try:
            PropertyDict[sensor_setting_property_id]._build(sensor_setting_raw, stream, context, path)
        except KeyError:
            stream_write(stream, bytes(sensor_setting_raw))

        return obj


class _SensorSettingSet(SensorSettingRawMixin, Construct):
    subcon = Struct(
        "sensor_property_id" / SensorPropertyId,
        "sensor_setting_property_id" / SensorPropertyId,
        PropertyValue,
    )

    def _parse(self, stream, context, path):
        obj = SensorSettingGet._parse(stream, context, path)

        sensor_setting_property_id = obj.pop("sensor_setting_property_id")
        return self._parse_sensor_setting(stream, context, path, sensor_setting_property_id, **obj)

    def _build(self, obj, stream, context, path):
        SensorSettingGet._build(obj, stream, context, path)

        sensor_setting_property_id = obj["sensor_setting_property_id"]
        return self._build_sensor_setting(obj, stream, context, path, sensor_setting_property_id)


SensorSettingSet = _SensorSettingSet()


class _SensorSettingStatus(SensorSettingRawMixin, Construct):
    subcon = Struct(
        "sensor_property_id" / SensorPropertyId,
        "sensor_setting_property_id" / SensorPropertyId,
        "sensor_setting_access" / Int8ul,
        PropertyValue,
    )

    def _parse(self, stream, context, path):
        obj = Struct(Embedded(SensorSettingGet), "sensor_setting_access" / Int8ul)._parse(stream, context, path)

        sensor_setting_property_id = obj.pop("sensor_setting_property_id")
        return self._parse_sensor_setting(stream, context, path, sensor_setting_property_id, **obj)

    def _build(self, obj, stream, context, path):
        Struct(Embedded(SensorSettingGet), "sensor_setting_access" / Int8ul)._build(obj, stream, context, path)

        sensor_setting_property_id = obj["sensor_setting_property_id"]
        return self._build_sensor_setting(obj, stream, context, path, sensor_setting_property_id)


SensorSettingStatus = _SensorSettingStatus()

SensorSettingsStatus = Struct(
    Embedded(SensorSettingsGet),
    "sensor_setting_property_ids" / GreedyRange(SensorPropertyId)
)

SensorDescriptorMinimal = Struct(
    "sensor_property_id" / SensorPropertyId,
)

SensorDescriptorOptional = Struct(
    Embedded(SensorDescriptorMinimal),
    *DoubleKeyIndex("sensor_negative_tolerance", "sensor_positive_tolerance"),
    "sensor_sampling_funcion" / Int8ul,
    "sensor_measurement_period" / Int8ul,
    "sensor_update_interval" / Int8ul
)

SensorDescriptorStatusItem = NamedSelect(
    optional=SensorDescriptorOptional,
    minimal=SensorDescriptorMinimal
)

SensorDescriptorStatus = GreedyRange(SensorDescriptorStatusItem)

class _SensorData(SensorSettingRawMixin, Construct):
    subcon = Struct(
        "format" / Int8ul,
        "length" / Int16ul,
        "sensor_setting_property_id" / SensorPropertyId,
        PropertyValue,
    )

    def _parse(self, stream, context, path):
        setting_property_id = stream_read(stream, 2)

        format = setting_property_id[0] & 0x01

        if format:
            setting_property_id += stream_read(stream, 1)

            length = (setting_property_id[0] >> 1) + 1
            sensor_setting_property_id = setting_property_id[1] | setting_property_id[2] << 8
        else:
            length = ((setting_property_id[0] >> 1) & 0b1111) + 1
            sensor_setting_property_id = (setting_property_id[0] >> 5 & 0b111) | setting_property_id[1] << 3

        substream = io.BytesIO(stream.read(length))
        return self._parse_sensor_setting(substream, context, path, sensor_setting_property_id, format=format, length=length)

    def _build(self, obj, stream, context, path):
        sensor_setting_property_id = obj["sensor_setting_property_id"]

        format = obj["format"]
        length = obj["length"]

        if format:
            stream_write(stream, bytes([(length - 1) << 1 | 0x01]))
            stream_write(stream, sensor_setting_property_id.to_bytes(2, byteorder='little'))
        else:
            stream_write(stream, bytes([(length - 1) << 1 | (sensor_setting_property_id & 0b111) << 5]))
            stream_write(stream, bytes([sensor_setting_property_id >> 3]))

        return self._build_sensor_setting(obj, stream, context, path, sensor_setting_property_id)

SensorData = _SensorData()

SensorStatus = GreedyRange(SensorData)

# TODO: message not implemented due to somewhat complicated structure and lack of examples
# SensorColumnGet = Struct(
#     Embedded(SensorSettingsGet),
#     "raw_value_x" / PropertyValue
# )

# TODO: message not implemented due to somewhat complicated structure and lack of examples
# SensorSeriesGet = Struct(
#     Embedded(SensorSettingsGet),
#     "raw_value_y" / PropertyValue
# )

FastCadencePeriodDivisorAndTriggerType = EmbeddedBitStruct(
    "status_trigger_type" / BitsInteger(1),
    "fast_cadence_period_divisor" / BitsInteger(7)
)

UnitlessTriggerDelta = DefaultCountValidator(Int8ul, rounding=1, resolution=0.1)

TriggerDelta = Struct(
    Switch(
        this.status_trigger_type,
        {
            0: Struct(
                "status_trigger_delta_down" / PropertyValue,
                "status_trigger_delta_up" / PropertyValue
            ),
            1: Struct(
                "status_trigger_delta_down" / UnitlessTriggerDelta,
                "status_trigger_delta_up" / UnitlessTriggerDelta
            )
        }
    )
)

# TODO: message not implemented due to somewhat complicated structure and lack of examples
# SensorCadence = Struct(
#     "sensor_setting_property_id" / Int16ul,
#     *FastCadencePeriodDivisorAndTriggerType,
#     Embedded(TriggerDelta),
#     "status_min_interval" / ExprAdapter(Int16ul, lambda obj, ctx: pow(2, obj), lambda obj, ctx: log(obj, 2)),
#     "fast_cadence_low" / PropertyValue,
#     "fast_cadence_high" / PropertyValue
# )

SensorMessage = SwitchStruct(
    "opcode" / Opcode(SensorOpcode),
    "params" / Switch(
        this.opcode,
        {
            SensorOpcode.SENSOR_DESCRIPTOR_GET: SensorGet,
            SensorOpcode.SENSOR_DESCRIPTOR_STATUS: SensorDescriptorStatus,
            SensorOpcode.SENSOR_GET: SensorGet,
            SensorOpcode.SENSOR_STATUS: SensorStatus,
            # SensorOpcode.SENSOR_COLUMN_GET: SensorColumnGet,
            # SensorOpcode.SENSOR_COLUMN_STATUS: 0x00,
            # SensorOpcode.SENSOR_SERIES_GET: SensorSeriesGet,
            # SensorOpcode.SENSOR_SERIES_STATUS: 0x00,
        }
    )
)

SensorSetupMessage = SwitchStruct(
    "opcode" / Opcode(SensorSetupOpcode),
    "params" / Switch(
        this.opcode,
        {
            SensorSetupOpcode.SENSOR_CADENCE_GET: SensorGetOptional,
            # SensorSetupOpcode.SENSOR_CADENCE_SET: SensorCadence,
            # SensorSetupOpcode.SENSOR_CADENCE_SET_UNACKNOWLEDGED: SensorCadence,
            # SensorSetupOpcode.SENSOR_CADENCE_STATUS: SensorCadence,
            SensorSetupOpcode.SENSOR_SETTINGS_GET: SensorSettingsGet,
            SensorSetupOpcode.SENSOR_SETTINGS_STATUS: SensorSettingsStatus,
            SensorSetupOpcode.SENSOR_SETTING_GET: SensorSettingGet,
            SensorSetupOpcode.SENSOR_SETTING_SET: SensorSettingSet,
            SensorSetupOpcode.SENSOR_SETTING_SET_UNACKNOWLEDGED: SensorSettingSet,
            SensorSetupOpcode.SENSOR_SETTING_STATUS: SensorSettingStatus,
        }
    )
)
# fmt: on



from construct import Construct, Container, SizeofError, stream_read_entire, stream_write

from .config import ConfigMessage, ConfigOpcode
from .generic.battery import GenericBatteryMessage, GenericBatteryOpcode
from .generic.level import GenericLevelMessage, GenericLevelOpcode
from .generic.onoff import GenericOnOffMessage, GenericOnOffOpcode
from .health import HealthMessage, HealthOpcode
from .light.ctl import LightCTLMessage, LightCTLOpcode, LightCTLSetupMessage, LightCTLSetupOpcode
from .light.lightness import (
    LightLightnessMessage,
    LightLightnessOpcode,
    LightLightnessSetupMessage,
    LightLightnessSetupOpcode,
)
from .scene import SceneMessage, SceneOpcode
from .sensor import SensorMessage, SensorOpcode, SensorSetupMessage, SensorSetupOpcode
from .silvair.debug import DebugMessage, DebugOpcode
from .silvair.debugV2 import DebugV2Message, DebugV2Opcode
from .silvair.emergency_lighting_test import (
    EmergencyLightingTestServerMessage,
    EmergencyLightingTestServerOpcode,
)
from .silvair.gateway_config_server import GatewayConfigMessage, GatewayConfigServerOpcode
from .silvair.light_extended_controller import LightExtendedControllerMessage, LightExtendedControllerOpcode
from .silvair.network_diagnostic_server import (
    NetworkDiagnosticServerMessage,
    NetworkDiagnosticServerOpcode,
    NetworkDiagnosticSetupServerMessage,
    NetworkDiagnosticSetupServerOpcode,
)
from .silvair.rrule_scheduler import RRuleSchedulerMessage, RRuleSchedulerOpcode
from .time import TimeMessage, TimeOpcode
from .util import Opcode


class _AccessMessage(Construct):
    OPCODES = {  # noqa: RUF012
        ConfigOpcode: ConfigMessage,
        DebugOpcode: DebugMessage,
        GatewayConfigServerOpcode: GatewayConfigMessage,
        GenericOnOffOpcode: GenericOnOffMessage,
        GenericBatteryOpcode: GenericBatteryMessage,
        GenericLevelOpcode: GenericLevelMessage,
        LightLightnessOpcode: LightLightnessMessage,
        LightLightnessSetupOpcode: LightLightnessSetupMessage,
        LightCTLOpcode: LightCTLMessage,
        LightCTLSetupOpcode: LightCTLSetupMessage,
        HealthOpcode: HealthMessage,
        NetworkDiagnosticServerOpcode: NetworkDiagnosticServerMessage,
        NetworkDiagnosticSetupServerOpcode: NetworkDiagnosticSetupServerMessage,
        SceneOpcode: SceneMessage,
        SensorOpcode: SensorMessage,
        SensorSetupOpcode: SensorSetupMessage,
        LightExtendedControllerOpcode: LightExtendedControllerMessage,
        TimeOpcode: TimeMessage,
        DebugV2Opcode: DebugV2Message,
        RRuleSchedulerOpcode: RRuleSchedulerMessage,
        EmergencyLightingTestServerOpcode: EmergencyLightingTestServerMessage,
    }

    OPCODE = Opcode()

    def __init__(self):
        super().__init__()
        self._opcodes = {}
        for opcode_class, message in self.OPCODES.items():
            for opcode in opcode_class._value2member_map_.keys():
                self._opcodes[opcode] = opcode_class(opcode), message.compile()

    def _parse(self, stream, context, path):
        opcode = self.OPCODE._parse(stream, context, path)

        try:
            opcode, message = self._opcodes[opcode]
        except KeyError:
            return Container(opcode=opcode, params=stream_read_entire(stream))

        stream.seek(0)
        parsed = message._parse(stream, context, path)
        parsed.opcode = opcode
        return parsed

    def _build(self, obj, stream, context, path):
        opcode = obj["opcode"]

        try:
            if isinstance(opcode, str):
                opcode, message = next(v for k, v in self._opcodes.items() if v[0].name == opcode)
                obj["opcode"] = opcode
            else:
                opcode, message = self._opcodes[opcode]
        except KeyError:
            Opcode()._build(opcode, stream, context, path)
            stream_write(stream, obj["params"])
            return obj

        return message._build(obj, stream, context, path)

    def _sizeof(self, context, path):
        raise SizeofError


AccessMessage = _AccessMessage()
